@@ -35,7 +35,7 @@ import           Data.Functor.Classes (Eq1(..), Ord1(..), Read1(..), Show1(..))
 import           Data.Functor.Classes (Eq2(..), Ord2(..), Read2(..), Show2(..))
 # endif
 #endif
-import           Data.List
+import qualified Data.List as List
 import qualified Data.Map as Map
 import           Data.Map (Map)
 import           Data.Maybe
@@ -66,6 +66,7 @@ import           Data.Char (isSymbol, ord)
 #endif
 
 import           Language.Haskell.TH.Datatype
+import           Language.Haskell.TH.Datatype.TyVarBndr
 import           Language.Haskell.TH.Lib
 import           Language.Haskell.TH.Ppr (pprint)
 import           Language.Haskell.TH.Syntax
@@ -375,7 +376,7 @@ buildTypeInstance cRep tyConName dataCxt varTysOrig variant = do
         --   instance C (Fam [Char])
         remainingTysOrigSubst :: [Type]
         remainingTysOrigSubst =
-          map (substNamesWithKindStar (union droppedKindVarNames kvNames'))
+          map (substNamesWithKindStar (List.union droppedKindVarNames kvNames'))
             $ take remainingLength varTysOrig
 
         isDataFamily :: Bool
@@ -813,14 +814,12 @@ newNameList :: String -> Int -> Q [Name]
 newNameList prefix n = mapM (newName . (prefix ++) . show) [1..n]
 
 -- | Extracts the kind from a TyVarBndr.
-tvbKind :: TyVarBndr -> Kind
-tvbKind (PlainTV  _)   = starK
-tvbKind (KindedTV _ k) = k
+tvbKind :: TyVarBndr_ flag -> Kind
+tvbKind = elimTV (\_ -> starK) (\_ k -> k)
 
 -- | Convert a TyVarBndr to a Type.
-tvbToType :: TyVarBndr -> Type
-tvbToType (PlainTV n)    = VarT n
-tvbToType (KindedTV n k) = SigT (VarT n) k
+tvbToType :: TyVarBndr_ flag -> Type
+tvbToType = elimTV VarT (\n k -> SigT (VarT n) k)
 
 -- | Applies a typeclass constraint to a type.
 applyClass :: Name -> Name -> Pred
@@ -970,7 +969,7 @@ predMentionsName (EqualP t1 t2) names = mentionsName t1 names || mentionsName t2
 
 -- | Construct a type via curried application.
 applyTy :: Type -> [Type] -> Type
-applyTy = foldl' AppT
+applyTy = List.foldl' AppT
 
 -- | Fully applies a type constructor to its type variables.
 applyTyCon :: Name -> [Type] -> Type
@@ -1042,10 +1041,9 @@ tag2ConExpr :: Type -> Q Exp
 tag2ConExpr ty = do
     iHash  <- newName "i#"
     ty' <- freshenType ty
-    let tvbs = avoidTypeInType $ freeVariablesWellScoped [ty']
     lam1E (conP iHashDataName [varP iHash]) $
         varE tagToEnumHashValName `appE` varE iHash
-            `sigE` return (ForallT tvbs [] ty')
+            `sigE` return (quantifyType ty')
             -- tagToEnum# is a hack, and won't typecheck unless it's in the
             -- immediate presence of a type ascription like so:
             --
@@ -1058,39 +1056,6 @@ tag2ConExpr ty = do
             -- (using th-abstraction's quantifyType function). Also make sure
             -- to freshen the bound type variables to avoid shadowed variable
             -- warnings on old versions of GHC when -Wall is enabled.
-  where
-    -- Somewhat annoyingly, it's possible to generate code that requires
-    -- TypeInType (on old versions of GHC) for data types which didn't require
-    -- TypeInType to define. To avoid users having to turn on more language
-    -- extensions than is necessary, we filter out all kind variable binders.
-    -- Fortunately, old versions of GHC are quite alright with implicitly
-    -- quantifying kind variables, even in the type of a SigE.
-    --
-    -- This is rather tiresome, and while writing this function, I debated
-    -- whether to just forget about this nonsense and require users to
-    -- enable TypeInType to use the generated code. Alas, that would entail
-    -- a breaking change, so I decided against it at the time. If we ever make
-    -- some breaking change in the future, however, this would be at the top
-    -- of the list of things that I'd rip out.
-    avoidTypeInType :: [TyVarBndr] -> [TyVarBndr]
-#if __GLASGOW_HASKELL__ >= 806
-    avoidTypeInType = id
-#else
-    avoidTypeInType = go . map attachFreeKindVars
-      where
-        attachFreeKindVars :: TyVarBndr -> (TyVarBndr, [Name])
-        attachFreeKindVars tvb = (tvb, freeVariables (tvKind tvb))
-
-        go :: [(TyVarBndr, [Name])] -> [TyVarBndr]
-        go [] = []
-        go ((tvb, _):tvbsAndFVs)
-          | any (\(_, kindVars) -> tvName tvb `elem` kindVars) tvbsAndFVs
-          = tvbs'
-          | otherwise
-          = tvb:tvbs'
-          where
-            tvbs' = go tvbsAndFVs
-#endif
 
 primOrdFunTbl :: Map Name (Name, Name, Name, Name, Name)
 primOrdFunTbl = Map.fromList
@@ -1155,6 +1120,20 @@ primOrdFunTbl = Map.fromList
                            , geWord16HashValName
                            , gtWord16HashValName
                            ))
+#endif
+#if MIN_VERSION_base(4,16,0)
+    , (int32HashTypeName,  ( ltInt32HashValName
+                           , leInt32HashValName
+                           , eqInt32HashValName
+                           , geInt32HashValName
+                           , gtInt32HashValName
+                           ))
+    , (word32HashTypeName, ( ltWord32HashValName
+                           , leWord32HashValName
+                           , eqWord32HashValName
+                           , geWord32HashValName
+                           , gtWord32HashValName
+                           ))
 #endif
     ]
 
@@ -1583,6 +1562,9 @@ maxBoundValName = mkNameG_v "base" "GHC.Enum" "maxBound"
 minusIntHashValName :: Name
 minusIntHashValName = mkNameG_v "ghc-prim" "GHC.Prim" "-#"
 
+neqIntHashValName :: Name
+neqIntHashValName = mkNameG_v "ghc-prim" "GHC.Prim" "/=#"
+
 parenValName :: Name
 parenValName = mkNameG_v "base" "GHC.Read" "paren"
 
@@ -2146,18 +2128,6 @@ eqWord8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "eqWord8#"
 eqWord16HashValName :: Name
 eqWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "eqWord16#"
 
-extendInt8HashValName :: Name
-extendInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "extendInt8#"
-
-extendInt16HashValName :: Name
-extendInt16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "extendInt16#"
-
-extendWord8HashValName :: Name
-extendWord8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "extendWord8#"
-
-extendWord16HashValName :: Name
-extendWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "extendWord16#"
-
 geInt8HashValName :: Name
 geInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "geInt8#"
 
@@ -2185,9 +2155,41 @@ gtWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "gtWord16#"
 int8HashTypeName :: Name
 int8HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Int8#"
 
+int8ToIntHashValName :: Name
+int8ToIntHashValName = mkNameG_v "ghc-prim" "GHC.Prim"
+# if MIN_VERSION_base(4,16,0)
+  "int8ToInt#"
+# else
+  "extendInt8#"
+# endif
+
 int16HashTypeName :: Name
 int16HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Int16#"
 
+int16ToIntHashValName :: Name
+int16ToIntHashValName = mkNameG_v "ghc-prim" "GHC.Prim"
+# if MIN_VERSION_base(4,16,0)
+  "int16ToInt#"
+# else
+  "extendInt16#"
+# endif
+
+intToInt8HashValName :: Name
+intToInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim"
+# if MIN_VERSION_base(4,16,0)
+  "intToInt8#"
+# else
+  "narrowInt8#"
+# endif
+
+intToInt16HashValName :: Name
+intToInt16HashValName = mkNameG_v "ghc-prim" "GHC.Prim"
+# if MIN_VERSION_base(4,16,0)
+  "intToInt16#"
+# else
+  "narrowInt16#"
+# endif
+
 leInt8HashValName :: Name
 leInt8HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "leInt8#"
 
@@ -2215,6 +2217,88 @@ ltWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "ltWord16#"
 word8HashTypeName :: Name
 word8HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Word8#"
 
+word8ToWordHashValName :: Name
+word8ToWordHashValName = mkNameG_v "ghc-prim" "GHC.Prim"
+# if MIN_VERSION_base(4,16,0)
+  "word8ToWord#"
+# else
+  "extendWord8#"
+# endif
+
 word16HashTypeName :: Name
 word16HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Word16#"
+
+word16ToWordHashValName :: Name
+word16ToWordHashValName = mkNameG_v "ghc-prim" "GHC.Prim"
+# if MIN_VERSION_base(4,16,0)
+  "word16ToWord#"
+# else
+  "extendWord16#"
+# endif
+
+wordToWord8HashValName :: Name
+wordToWord8HashValName = mkNameG_v "ghc-prim" "GHC.Prim"
+# if MIN_VERSION_base(4,16,0)
+  "wordToWord8#"
+# else
+  "narrowWord8#"
+# endif
+
+wordToWord16HashValName :: Name
+wordToWord16HashValName = mkNameG_v "ghc-prim" "GHC.Prim"
+# if MIN_VERSION_base(4,16,0)
+  "wordToWord16#"
+# else
+  "narrowWord16#"
+# endif
+#endif
+
+#if MIN_VERSION_base(4,16,0)
+eqInt32HashValName :: Name
+eqInt32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "eqInt32#"
+
+eqWord32HashValName :: Name
+eqWord32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "eqWord32#"
+
+geInt32HashValName :: Name
+geInt32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "geInt32#"
+
+geWord32HashValName :: Name
+geWord32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "geWord32#"
+
+gtInt32HashValName :: Name
+gtInt32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "gtInt32#"
+
+gtWord32HashValName :: Name
+gtWord32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "gtWord32#"
+
+int32HashTypeName :: Name
+int32HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Int32#"
+
+int32ToIntHashValName :: Name
+int32ToIntHashValName = mkNameG_v "ghc-prim" "GHC.Prim" "int32ToInt#"
+
+intToInt32HashValName :: Name
+intToInt32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "intToInt32#"
+
+leInt32HashValName :: Name
+leInt32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "leInt32#"
+
+leWord32HashValName :: Name
+leWord32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "leWord32#"
+
+ltInt32HashValName :: Name
+ltInt32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "ltInt32#"
+
+ltWord32HashValName :: Name
+ltWord32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "ltWord32#"
+
+word32HashTypeName :: Name
+word32HashTypeName = mkNameG_tc "ghc-prim" "GHC.Prim" "Word32#"
+
+word32ToWordHashValName :: Name
+word32ToWordHashValName = mkNameG_v "ghc-prim" "GHC.Prim" "word32ToWord#"
+
+wordToWord32HashValName :: Name
+wordToWord32HashValName = mkNameG_v "ghc-prim" "GHC.Prim" "wordToWord32#"
 #endif
