@@ -1,4 +1,9 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE TupleSections #-}
+
+#if __GLASGOW_HASKELL__ >= 800
+{-# LANGUAGE TemplateHaskellQuotes #-}
+#endif
 
 {-|
 Module:      Data.Deriving.Via.Internal
@@ -23,12 +28,16 @@ module Data.Deriving.Via.Internal where
 import           Control.Monad ((<=<), unless)
 
 import           Data.Deriving.Internal
+import qualified Data.List as L
 import qualified Data.Map as M
 import           Data.Map (Map)
 import           Data.Maybe (catMaybes)
 
+import           GHC.Exts (Any)
+
 import           Language.Haskell.TH
 import           Language.Haskell.TH.Datatype
+import           Language.Haskell.TH.Datatype.TyVarBndr
 
 -------------------------------------------------------------------------------
 -- Code generation
@@ -45,12 +54,11 @@ $('deriveGND' [t| forall a. 'Eq' a => 'Eq' (Foo a) |])
 deriveGND :: Q Type -> Q [Dec]
 deriveGND qty = do
   ty <- qty
-  let (instanceTvbs, instanceCxt, instanceTy) = decomposeType ty
+  let (_instanceTvbs, instanceCxt, instanceTy) = decomposeType ty
   instanceTy' <- (resolveTypeSynonyms <=< resolveInfixT) instanceTy
   decs <- deriveViaDecs instanceTy' Nothing
-  let instanceHeader = ForallT instanceTvbs instanceCxt instanceTy
-  (:[]) `fmap` instanceD (return [])
-                         (return instanceHeader)
+  (:[]) `fmap` instanceD (return instanceCxt)
+                         (return instanceTy)
                          (map return decs)
 
 {- | Generates an instance for a type class by emulating the behavior of the
@@ -70,7 +78,7 @@ correctly across all the types being used (e.g., to make sure that the same
 deriveVia :: Q Type -> Q [Dec]
 deriveVia qty = do
   ty <- qty
-  let (instanceTvbs, instanceCxt, viaApp) = decomposeType ty
+  let (_instanceTvbs, instanceCxt, viaApp) = decomposeType ty
   viaApp' <- (resolveTypeSynonyms <=< resolveInfixT) viaApp
   (instanceTy, viaTy)
     <- case unapplyTy viaApp' of
@@ -82,10 +90,17 @@ deriveVia qty = do
                 , "\tThe ‘Via‘ type must be used, e.g."
                 , "\t[t| forall a. C (T a) `Via` V a |]"
                 ]
-  decs <- deriveViaDecs instanceTy (Just viaTy)
-  let instanceHeader = ForallT instanceTvbs instanceCxt instanceTy
-  (:[]) `fmap` instanceD (return [])
-                         (return instanceHeader)
+  -- This is a stronger requirement than what GHC's implementation of
+  -- DerivingVia imposes, but due to Template Haskell restrictions, we
+  -- currently can't do better. See #27.
+  let viaTyFVs           = freeVariables viaTy
+      otherFVs           = concat [freeVariables instanceCxt, freeVariables instanceTy]
+      floatingViaTyFVs   = viaTyFVs L.\\ otherFVs
+      floatingViaTySubst = M.fromList $ map (, ConT ''Any) floatingViaTyFVs
+      viaTy'             = applySubstitution floatingViaTySubst viaTy
+  decs <- deriveViaDecs instanceTy (Just viaTy')
+  (:[]) `fmap` instanceD (return instanceCxt)
+                         (return instanceTy)
                          (map return decs)
 
 deriveViaDecs :: Type       -- ^ The instance head (e.g., @Eq (Foo a)@)
@@ -133,7 +148,7 @@ deriveViaDecs instanceTy mbViaTy = do
         _ -> fail $ "Not a type class: " ++ pprint clsTy
     _ -> fail $ "Malformed instance: " ++ pprint instanceTy
 
-deriveViaDecs' :: Name -> [TyVarBndr] -> [Type] -> Type -> Dec -> Q (Maybe [Dec])
+deriveViaDecs' :: Name -> [TyVarBndrUnit] -> [Type] -> Type -> Dec -> Q (Maybe [Dec])
 deriveViaDecs' clsName clsTvbs clsArgs repTy dec = do
     let numExpectedArgs = length clsTvbs
         numActualArgs   = length clsArgs
@@ -172,7 +187,7 @@ deriveViaDecs' clsName clsTvbs clsArgs repTy dec = do
 
     go _ = return Nothing
 
-mkCoerceClassMethEqn :: [TyVarBndr] -> [Type] -> Type -> Type -> (Type, Type)
+mkCoerceClassMethEqn :: [TyVarBndrUnit] -> [Type] -> Type -> Type -> (Type, Type)
 mkCoerceClassMethEqn clsTvbs clsArgs repTy methTy
   = ( applySubstitution rhsSubst methTy
     , applySubstitution lhsSubst methTy
@@ -181,7 +196,7 @@ mkCoerceClassMethEqn clsTvbs clsArgs repTy methTy
     lhsSubst = zipTvbSubst clsTvbs clsArgs
     rhsSubst = zipTvbSubst clsTvbs $ changeLast clsArgs repTy
 
-zipTvbSubst :: [TyVarBndr] -> [Type] -> Map Name Type
+zipTvbSubst :: [TyVarBndr_ flag] -> [Type] -> Map Name Type
 zipTvbSubst tvbs = M.fromList . zipWith (\tvb ty -> (tvName tvb, ty)) tvbs
 
 -- | Replace the last element of a list with another element.
@@ -199,7 +214,7 @@ stripOuterForallT (ForallT _ _ ty) = ty
 #endif
 stripOuterForallT ty               = ty
 
-decomposeType :: Type -> ([TyVarBndr], Cxt, Type)
+decomposeType :: Type -> ([TyVarBndrSpec], Cxt, Type)
 decomposeType (ForallT tvbs ctxt ty) = (tvbs, ctxt, ty)
 decomposeType ty                     = ([],   [],   ty)
 
