@@ -52,7 +52,7 @@ module Text.Show.Deriving.Internal (
     ) where
 
 import           Data.Deriving.Internal
-import           Data.List
+import qualified Data.List as List
 import qualified Data.Map as Map
 import           Data.Map (Map)
 import           Data.Maybe (fromMaybe)
@@ -398,7 +398,7 @@ makeShowForCon p sClass opts tvMap
        then do
          let showArgs       = zipWith (makeShowForArg 0 sClass opts conName tvMap) argTys' args
              parenCommaArgs = (varE showCharValName `appE` charE '(')
-                              : intersperse (varE showCharValName `appE` charE ',') showArgs
+                              : List.intersperse (varE showCharValName `appE` charE ',') showArgs
              mappendArgs    = foldr (`infixApp` varE composeValName)
                                     (varE showCharValName `appE` charE ')')
                                     parenCommaArgs
@@ -692,34 +692,46 @@ primShowTbl = Map.fromList
                              })
 #if MIN_VERSION_base(4,13,0)
     , (int8HashTypeName,   PrimShow
-                             { primShowBoxer      = appE (conE iHashDataName) . appE (varE extendInt8HashValName)
+                             { primShowBoxer      = appE (conE iHashDataName) . appE (varE int8ToIntHashValName)
                              , primShowPostfixMod = oneHashE
-                             , primShowConv       = mkNarrowE "narrowInt8#"
+                             , primShowConv       = mkNarrowE intToInt8HashValName
                              })
     , (int16HashTypeName,  PrimShow
-                             { primShowBoxer      = appE (conE iHashDataName) . appE (varE extendInt16HashValName)
+                             { primShowBoxer      = appE (conE iHashDataName) . appE (varE int16ToIntHashValName)
                              , primShowPostfixMod = oneHashE
-                             , primShowConv       = mkNarrowE "narrowInt16#"
+                             , primShowConv       = mkNarrowE intToInt16HashValName
                              })
     , (word8HashTypeName,  PrimShow
-                             { primShowBoxer      = appE (conE wHashDataName) . appE (varE extendWord8HashValName)
+                             { primShowBoxer      = appE (conE wHashDataName) . appE (varE word8ToWordHashValName)
                              , primShowPostfixMod = twoHashE
-                             , primShowConv       = mkNarrowE "narrowWord8#"
+                             , primShowConv       = mkNarrowE wordToWord8HashValName
                              })
     , (word16HashTypeName, PrimShow
-                             { primShowBoxer      = appE (conE wHashDataName) . appE (varE extendWord16HashValName)
+                             { primShowBoxer      = appE (conE wHashDataName) . appE (varE word16ToWordHashValName)
                              , primShowPostfixMod = twoHashE
-                             , primShowConv       = mkNarrowE "narrowWord16#"
+                             , primShowConv       = mkNarrowE wordToWord16HashValName
+                             })
+#endif
+#if MIN_VERSION_base(4,16,0)
+    , (int32HashTypeName,  PrimShow
+                             { primShowBoxer      = appE (conE iHashDataName) . appE (varE int32ToIntHashValName)
+                             , primShowPostfixMod = oneHashE
+                             , primShowConv       = mkNarrowE intToInt32HashValName
+                             })
+    , (word32HashTypeName, PrimShow
+                             { primShowBoxer      = appE (conE wHashDataName) . appE (varE word32ToWordHashValName)
+                             , primShowPostfixMod = twoHashE
+                             , primShowConv       = mkNarrowE wordToWord32HashValName
                              })
 #endif
     ]
 
 #if MIN_VERSION_base(4,13,0)
-mkNarrowE :: String -> Q Exp -> Q Exp
-mkNarrowE narrowStr e =
+mkNarrowE :: Name -> Q Exp -> Q Exp
+mkNarrowE narrowName e =
   foldr (`infixApp` varE composeValName)
         (varE showCharValName `appE` charE ')')
-        [ varE showStringValName `appE` stringE ('(':narrowStr ++ " ")
+        [ varE showStringValName `appE` stringE ('(':nameBase narrowName ++ " ")
         , e
         ]
 #endif
