@@ -5,6 +5,7 @@
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE TypeOperators #-}
 #if __GLASGOW_HASKELL__ >= 708
 {-# LANGUAGE EmptyCase #-}
 #endif
@@ -25,6 +26,9 @@ import Data.Deriving
 import GHC.Exts ( Char#, Double#, Float#, Int#, Word#
 #if MIN_VERSION_base(4,13,0)
                 , Int8#, Int16#, Word8#, Word16#
+#endif
+#if MIN_VERSION_base(4,16,0)
+                , Int32#, Word32#
 #endif
                 )
 
@@ -53,6 +57,10 @@ data TyCon# a b = TyCon# {
   , tcWord8#  :: Word8#
   , tcWord16# :: Word16#
 #endif
+#if MIN_VERSION_base(4,16,0)
+  , tcInt32#  :: Int32#
+  , tcWord32# :: Word32#
+#endif
 }
 
 data TyCon2 a b c d where
@@ -94,6 +102,10 @@ data instance TyFamily# a b = TyFamily# {
   , tfWord8#  :: Word8#
   , tfWord16# :: Word16#
 #endif
+#if MIN_VERSION_base(4,16,0)
+  , tfInt32#  :: Int32#
+  , tfWord32# :: Word32#
+#endif
 }
 
 data family TyFamily2 w x y z :: *
